# -*- coding: utf-8 -*-

from os import listdir
from os.path import join, exists
import json
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import h5py
from IPython.display import display
from warnings import warn
from .util import dict_val_to_list, arg_to_list, set_to_several_objects
import pyberries as pyb

class DatasetPool():
    
    def __init__(self,
                 path,
                 dsList,
                 groups=[],
                 metadata={},
                 filters={},
                 preprocessing={}
                 ):
        
        path = arg_to_list(path)
        dsList = arg_to_list(dsList)
        self.path = path
        self.dsList = dsList
        self.table = dict()
        self.parent = dict()
        self.channel = dict()
        preprocessing = dict_val_to_list(preprocessing)

        for i, ds in enumerate(dsList):
            ds_path = path[0] if len(path) == 1 else path[i]
            assert exists(ds_path), f'Bacmman folder not found: {ds_path}'
            assert exists(join(ds_path, ds)), f'Dataset {ds} not found.\nMaybe looking for {" or ".join([x for x in listdir(ds_path) if x.startswith(ds[0:6])])}?'
            assert (len(groups) >= len(dsList)) or not groups, 'If groups are provided, one group should be defined per dataset'

            # Load configuration
            cf = join(ds_path, ds, f"{ds}_config.json")
            with open(cf, errors='ignore') as f:
                conf = json.load(f)
                object_class_names = [c["name"] for c in conf["structures"]["list"]]
                object_parents = [c["segmentationParent"] for c in conf["structures"]["list"]]
                channel_images = [c["channelImage"] for c in conf["structures"]["list"]]

            # Load data tables
            for j, obj in enumerate(object_class_names):
                df_in = pd.read_csv(join(ds_path, ds, f"{ds}_{j}.csv"), sep=';', low_memory=False)
                df_in['Dataset'] = ds
                df_in['Group'] = groups[i] if groups else i
                df_in[['Dataset', 'Group']] = df_in[['Dataset', 'Group']].astype('category')
                if obj in preprocessing.keys():
                    if len(preprocessing[obj]) == 1:
                        df_in = df_in.transform(preprocessing[obj][0])
                    elif len(preprocessing[obj]) > 1:
                        assert len(preprocessing[obj]) == len(self.dsList), 'If multiple pre-processings are provided, there should be one per dataset'
                        if preprocessing[obj][i]:
                            df_in = df_in.transform(preprocessing[obj][i])
                self.table[obj] = df_in if obj not in self.table.keys() else pd.concat([self.table[obj], df_in], axis=0)
                self.parent[obj] = object_class_names[object_parents[j][0]] if object_parents[j] else None
                self.channel[obj] = channel_images[j][0]
            print(f'Dataset {ds}: loaded objects {object_class_names}')
        self.objects = self.table.keys()
        for data in self.table.values():
            data.reset_index(drop=True, inplace=True)

        self.add_metadata(metadata)
        self.apply_filters(filters)

    def describe(self, agg, include='all'):
        include = arg_to_list(include)
        for obj, df in self.table.items():
            if include != ['all']:
                df = df.filter(items=include + ['Dataset', 'Group'])
            print(f'{obj}')
            if df.empty:
                print('No data')
            else:
                df1 = (df
                    .groupby('Dataset', sort=False)
                    .agg(nObjects = ('Dataset', 'count'))
                    )
                df2 =(df
                    .loc[:, ~df.columns.isin(['Position','PositionIdx','Indices','Frame','Idx','Time','Group'])]
                    .groupby('Dataset', sort=False)
                    .agg(agg, numeric_only=True)
                    )
                if isinstance(agg, list):
                    df2.columns = df2.columns.map(' ('.join) + ')'
                df3 = df[['Group', 'Dataset']].drop_duplicates().set_index('Dataset')
                df_out = df3.join([df1, df2])
                display(df_out)
    
    def add_metadata(self, metadata):
        metadata = set_to_several_objects(metadata, self.objects)
        metadata = dict_val_to_list(metadata)
        for obj in metadata.keys():
            if not self.table[obj].empty:
                # channel = self.channel[obj]
                channel = 0 # Other channels often miss metadata information so maybe better to always use this one?
                df = self.table[obj]
                for var in metadata[obj]:
                    df[var.replace(' ', '_')] = 0
                    if var == 'DateTime':
                        df['TimeDelta'] = 0
                        df['Time_min'] = 0
                    for i, ds in enumerate(self.dsList):
                        ds_path = self.path[0] if len(self.path) == 1 else self.path[i]
                        for pos in df.loc[df['Dataset'] == ds].Position.unique():
                            # Open metadata file
                            with open(join(ds_path, ds, 'SourceImageMetadata', f'{pos}_c{channel}.txt')) as f:
                                value = next((line for line in f if var in line), None)
                                # Add metadata value to the current dataset and position
                                df.loc[(df['Position'] == pos) & (df['Dataset'] == ds), var.replace(' ', '_')] = value.split('=')[-1][:-1] # Remove line break (last character)
                        if 'DateTime' in df.columns:
                            df.loc[df['Dataset'] == ds] = (df.loc[df['Dataset'] == ds]
                                                            .assign(DateTime = lambda df: pd.to_datetime(df.DateTime, format='%Y%m%d %H:%M:%S.%f'),
                                                                    TimeDelta = lambda df: df.DateTime - df.DateTime.iloc[0],
                                                                    Time_min = lambda df: df.TimeDelta.dt.total_seconds().div(60)
                                                                    )
                                                            )
                self.table[obj] = df
    
    def apply_filters(self, filters):
        filters = set_to_several_objects(filters, self.objects)
        filters = dict_val_to_list(filters)
        for obj, filter in filters.items():
            if len(filter) == 1:
                self.table[obj] = self.table[obj].query(filter[0])
            elif len(filter) > 1:
                assert len(filter) == len(self.dsList), 'If multiple filters are provided, there should be one per dataset'
                df = pd.DataFrame()
                i=0
                for _, data in self.table[obj].groupby('Dataset', sort=False):
                    if filter[i]:
                        df = pd.concat([df, data.query(filter[i])], axis=0)
                    else:
                        df = pd.concat([df, data], axis=0)
                    i+=1
                self.table[obj] = df
            # Propagate filters to children (if any)
            for child, parent in self.parent.items():
                if parent == obj:
                    self.propagate_filters(parent, child)

    def rename_object(self, rename:dict):
        for old_name, new_name in rename.items():
            if new_name in self.objects:
                self.table[new_name] = pd.concat([self.table[new_name], self.table[old_name]], axis=0).reset_index(drop=True, inplace=True)
            else:
                self.table[new_name] = self.table[old_name]
                self.channel[new_name] = self.channel[old_name]
                self.parent[new_name] = self.parent[old_name]
            del self.table[old_name]
            del self.channel[old_name]
            del self.parent[old_name]
            self.objects = self.table.keys()

    def get_parent_indices(self, obj:str, indices:str='Indices', newcol:str='ParentIndices'):
        self.table[obj][newcol] = (self.table[obj]
                                            [indices]
                                            .str.split('-', expand=True)
                                            .iloc[:,:-1]
                                            .agg('-'.join, axis=1)
                                            )
    
    def get_idx(self, obj:str, idx:int=0, indices:str='Indices', newcol:str='ParentIdx'):
        self.table[obj][newcol] = (self.table[obj]
                                            [indices]
                                            .str.split('-', expand=True)
                                            .iloc[:,idx]
                                            .astype('int64')
                                            )
        
    def fuse_columns(self, obj:str, cols:list=[], new:str='new_col'):
        self.table[obj][new] = self.table[obj][cols].astype('str').agg('-'.join, axis=1)

    def normalise(self, object_name:str, col:str='', normalise_by:str='', new_col:str=''):
        if new_col in self.table[object_name].columns:
            self.table[object_name].drop(columns=new_col, inplace=True)
        self.table[object_name] = (self.table[object_name]
                                        .assign(new_tmp = lambda df: df[col] / df[normalise_by])
                                        .rename(columns={'new_tmp':new_col})
                                        )
    
    def propagate_filters(self, parent:str, child:str):
        self.get_parent_indices(obj=child)
        self.table[child] = (self.table[child]
                            .merge(self.table[parent][['Dataset', 'PositionIdx', 'Indices']], suffixes=(None, '_tmp'), left_on = ['Dataset', 'PositionIdx', 'ParentIndices'], right_on=['Dataset', 'PositionIdx', 'Indices'])
                            .transform(lambda df: df.loc[:, ~df.columns.str.contains('_tmp')])
                            )
    
    def add_background(self, object_name:str, h5filenames:list, selection_name:str='Viewfield', raw_keyword:str='raw', seg_keyword:str='cells'):
        df = self.table[object_name]
        for i, ds in enumerate(self.dsList):
            ds_path = self.path[0] if len(self.path) == 1 else self.path[i]
            h5filename = h5filenames[0] if len(h5filenames) == 1 else h5filenames[i]
            h5file = join(ds_path, ds, h5filename)
            assert exists(h5file), 'h5 file not found'
            f = h5py.File(h5file, "r")
            for p in df.loc[df['Dataset'] == ds].Position.unique():
                # Get raw fluorescence image
                raw = np.array(f[f'{selection_name}/{ds}/{p}/{raw_keyword}'])   
                # Get segmentation mask for current image
                cells_seg = np.array(f[f'{selection_name}/{ds}/{p}/{seg_keyword}'])
                for im in range(raw.shape[0]):
                    df.loc[(df.Dataset == ds) & (df.Position == p) & (df.Frame == im), ('Background')] = np.mean(raw[im, cells_seg[im,] == 0])
            f.close()
        self.table[object_name] = df

    def add_from_parent(self, object_name:str, col:list=[]):
        parent = self.parent[object_name]
        self.get_parent_indices(obj=object_name)
        df = self.table[object_name]
        for c in col:
            df = (df
                .merge(self.table[parent][['PositionIdx', 'Indices', c]], suffixes=(None, '_tmp'), left_on = ['PositionIdx', 'ParentIndices'], right_on=['PositionIdx', 'Indices'])
                .transform(lambda df: df.loc[:, ~df.columns.str.contains('_tmp')])
                )
        self.table[object_name] = df

    def add_columns(self, object_name:str, metrics, **kwargs):
        df = self.table[object_name]
        for m in metrics:
            if m == 'Heatmap':
                df = pyb.metrics.heatmap_metrics(df, **kwargs)
            elif m == 'is_col_larger':
                df = pyb.metrics.is_col_larger(df, **kwargs)
            elif m == 'bin_column':
                df = pyb.metrics.bin_column(df, **kwargs)
            elif m == 'Dapp':
                df = pyb.metrics.tracking_Dapp(df, **kwargs)
            else:
                ValueError(f'Metric "{m}" not found')
        self.table[object_name] = df

    def get_timeseries(self, object_name:str, metric, **kwargs):
        df_in = self.table[object_name]
        self.timeseries = dict()
        if metric == 'SpineLength':
            df_out = pyb.metrics.Cell_length(df_in, **kwargs)
        elif metric == 'ObjectCount':
            df_out = pyb.metrics.Object_count(df_in, **kwargs)
        elif metric == 'ObjectClass':
            df_out = pyb.metrics.Object_classifier(df_in, **kwargs)
        elif metric == 'Intensity':
            df_out = pyb.metrics.Object_intensity(df_in, **kwargs)
        elif metric == 'Quantile':
            df_out = pyb.metrics.Column_quantile(df_in, **kwargs)
        elif metric == 'Aggregation':
            df_out = pyb.metrics.Column_aggregation(df_in, **kwargs)
        elif metric == 'Fluo_intensity':
            df_out = pyb.metrics.Fluo_intensity(df_in, **kwargs)
        else:
            ValueError(f'Metric "{metric}" not found')
        self.timeseries[object_name] = df_out

    def plot_preset(self, preset:str, object_name:str='', timeseries:bool=False, drop_duplicates_by=[], return_axes:bool=False, **kwargs):
        if object_name:
            df_in = self.timeseries[object_name] if timeseries else self.table[object_name]
            if drop_duplicates_by:
                df_in = df_in.drop_duplicates(subset = drop_duplicates_by)
        _,ax = plt.subplots(dpi=130)
        if preset == 'histogram':
            ax = pyb.plots.plot_histogram(df_in, ax=ax, **kwargs)
        elif preset == 'histogram_fit':
            ax = pyb.plots.plot_histogram_fit(df_in, ax=ax, **kwargs)
        elif preset == 'bar':
            ax = pyb.plots.barplot(df_in, ax=ax, **kwargs)
        elif preset == 'line':
            ax = pyb.plots.lineplot(df_in, ax=ax, **kwargs)
        elif preset == 'scatter':
            ax = pyb.plots.scatterplot(df_in, ax=ax, **kwargs)
        elif preset == 'datapoints_and_mean':
            ax = pyb.plots.plot_datapoints_and_mean(df_in, dsList=self.dsList, ax=ax, **kwargs)
        elif preset == 'heatmap':
            ax = pyb.plots.plot_heatmap(df_in, ax=ax, **kwargs)
        elif preset == 'timeseries':
            ax = pyb.plots.plot_timeseries(df_in, ax=ax, **kwargs)
        elif preset == 'boxenplot':
            ax = pyb.plots.plot_boxenplot(df_in, ax=ax, **kwargs)
        elif preset == 'spot_tracks':
            lineage = kwargs.pop('lineage','')
            self.fuse_columns(obj=object_name, cols=['Idx','BacteriaLineage'], new='Track')
            if lineage:
                df = df_in.copy().query('BacteriaLineage == @lineage')
            else:
                df = df_in.copy()
            ax=pyb.plots.lineplot(df, hue='Track', sort=False, ax=ax, **kwargs)
        elif preset == 'rates_summary':
            ax = pyb.plots.plot_rates_summary(ax=ax, **kwargs)
        else:
            warn('Plot preset not found!')
        if return_axes: return ax